/*---------------------------------------------------------------*/
/* 8-pin SD audio WAV-player R0.04                 (C)ChaN, 2010 */
/*---------------------------------------------------------------*/
/*
  :
  SUT10=00             14CK + 1K (1024) CK + 4 ms, BOD enabled
  CKSEL3..0=0001       High Frequency PLL Clock Operating Modes
  BODLEVEL2...0=101    Brown-Out Detection = 2.7V
  RSTDISEL=0 -     HF  
  CKDIV8=1
  CKOUT=1
  DWEN=1
  EESAVE=0
  0 -  
*/
/*     asmfunc.S  Makefile:
  MODE=0 -  mo ( 8 )
  MODE=1 -  st ( 8 )
  MODE=2 -  hf ( 16 )
*/
//
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <string.h>
#include "diskio.h"
#include "pff.h"
//
# include <avr/eeprom.h>                 //    EEPROM
//
#ifndef MODE
#error Wrong make file.
#endif
#define FCC(c1,c2,c3,c4) (((DWORD)c4<<24)+((DWORD)c3<<16)+((WORD)c2<<8)+(BYTE)c1) /* FourCC */
//
//    GPIOR0 - General Purpose I/O Register 0, Address=0x11(+0x20)
#define GPIOR0 _SFR_IO8(0x11)
#define SAVE_EEPR 1                     //1-     EEPROM, 0-
//
/*---------------------------------------------------------*/
/* Work Area                                               */
/*---------------------------------------------------------*/
volatile BYTE FifoRi, FifoWi, FifoCt;       /* FIFO controls */
//FifoCt - FIFO data counter (  FIFO)
//FifoRi - FIFO read index
//FifoWi - FIFO write index
BYTE Buff[256];                             /* Wave output FIFO */
FATFS Fs;                                   /* File system object */
DIR Dir;                                    /* Directory object */
FILINFO Fno;                                /* File information - */
WORD rb;                                    /* Return value. Put this here to avoid bugs of avr-gcc */
EMPTY_INTERRUPT(WDT_vect);
unsigned char ButB=0;                       //    
unsigned int K_Trek=0;                      //  ,    
unsigned int N_File=0;                      //  
/*---   EEPROM---*/
unsigned int EEMEM Eindex;                  //   ( ): 0...65535
//unsigned int EEMEM ENtrek;                  //    : 0...65535 -  
//
/*---------------------------------------------------------*/
//   (12 )
static DWORD load_header (void)             /* 0:Invalid format, 1:I/O error, >1:Number of samples */
{ DWORD sz;
  if (pf_read(Buff, 12, &rb)) return 1;     /* Load file header (12 bytes) */
  if (rb != 12  ||  LD_DWORD(Buff+8) != FCC('W','A','V','E')) return 0;
  for (;;)
  { pf_read(Buff, 8, &rb);                  /* Get Chunk ID and size */
    if (rb != 8) return 0;
    sz = LD_DWORD(&Buff[4]);                /* Chunk size */
    switch (LD_DWORD(&Buff[0]))             /* FCC */
    { case FCC('f','m','t',' ') :           /* 'fmt ' chunk */
        if (sz > 100 || sz < 16) return 0;  //Check chunk size -     
        pf_read(Buff, sz, &rb);             /* Get content */
        if (rb != sz) return 0;
        if (Buff[0] != 1) return 0;         /* Check coding type (1) */
        if (Buff[2] != 1 && Buff[2] != 2) return 0; /* Check channels (1/2) */
        GPIOR0 = Buff[2];                   /* Get channel flag */
        if (Buff[14] != 8 && Buff[14] != 16) return 0; /* Check resolution (8/16) */
        GPIOR0 |= Buff[14];                 /* Resolution flag */
        OCR0A = (BYTE)(F_CPU/8/LD_WORD(&Buff[4]))-1; /* Sampling freq */
        break;
      case FCC('d','a','t','a') :           /* 'data' chunk (start to play) */
        return sz;
      case FCC('L','I','S','T') :           /* 'LIST' chunk (skip) */
      case FCC('f','a','c','t') :           /* 'fact' chunk (skip) */
        pf_lseek(Fs.fptr + sz);    /* pf_lseek(Offset)   /   .
                                      Offset (     )   
                                        .
                                    */
        break;
      default :                             /* Unknown chunk (error) */
        return 0;
    }
  }
  return 0;
}
//
/*---- PLAY----*/
static UINT play (const char *fn)
{ DWORD sz;
  FRESULT res;                              //,   pf_read()
  BYTE sw;                                  //  ">>"
  WORD btr;                                 //   (ByteToRead)
//
  unsigned int CntButt=0;                   //   
//
  if ((res = pf_open(fn)) == FR_OK)         //  *fn,   OK,  :
  { //     EEPROM
    #ifdef SAVE_EEPR
      eeprom_write_word(&Eindex, N_File);
    #endif
    //
    sz = load_header();                     //Load file header -   
    if (sz < 256) return (UINT)sz;
    if (!TCCR1)                             /* Enable audio out if not enabled */
    { PLLCSR = 0b00000110;                  /* High speed mode, asynchronous clock mode, PLL clock internal RC-oscillator for TC1.ck */
      GTCCR =  0b01100000;                  /* Enable TC1.OCB as PWM out (L-ch) - PWM mode  OCR1B, Clear the OC1B output line*/
      OCR1B = 128; OCR1A = 128;
      TCCR1 = MODE ? 0b01100001 : 0b00000001;/* Start TC1 with TC1.OCA is enabled as PWM out (R-ch) */
/* MODE=0 TCCR1=0b00000001 -  , CK
   MODE=1  2 TCCR1=0b01100001 - PWM mode  OCR1A  TC1, Clear the OC1A output line, CK
*/
      TCCR0A = 0b00000010;                  /* Enable TC0.ck = 2MHz as interval timer, CTC*/
      TCCR0B = 0b00000010;                  //clkI/O/8 (From prescaler)
      TIMSK = _BV(OCIE0A);                  //Timer/Counter0 Compare Match A interrupt is enabled
    }
    FifoCt = 0; FifoRi = 0; FifoWi = 0;     /* Reset FIFO */
    pf_read(0, 512 - (Fs.fptr % 512), &rb); /* Snip sector unaligned part -    */
    sz=sz-rb;                               //sz -= rb;
    sw = 1;                                 /* Button status flag -   PINB0*/
    do                                      /* Forward audio data */
    { btr = (sz > 1024) ? 1024 : (WORD)sz;
      res = pf_read(0, btr, &rb);           // btr   0- , rb     ()
      if (res != FR_OK || btr != rb) break;
      sz=sz-rb;                             //sz -= rb;
      /* Check button down and break on button down */
      sw=sw<<1;                             //sw <<= 1;
      if (bit_is_clear(PINB, 0) && ++sw == 1) break;//    8-  -    
/*      , sw=0,  ++sw==1 
     , sw=00000010,  ++sw=00000011=1 
*/
  //     ">>"
    #ifdef SAVE_EEPR
      if (bit_is_clear(PINB, 0))            //  
      { CntButt++;                          //     
        if (CntButt>=250)                   //    250 
        { ButB=1;                           //       
          break;                            //   
        }
      }
      else
      { ButB=0;                          //     
        CntButt=0;
      }
    #endif
  //
      wdt_reset();
    } while (rb==1024);                     //Repeat until all data read - ,     
  }
  while (FifoCt);                           //Wait for FIFO empty -    FIFO
  OCR1A = 128; OCR1B = 128;
  return res;
}
//
//
/*---- */
static void delay500 (void)
{ TCCR1 = 0; GTCCR = 0;                     /* Stop TC1 */
  TCCR0B = 0; TCCR0A = 0;                   /* Stop TC0 */
  wdt_reset();
  WDTCR = _BV(WDE) | _BV(WDIE) | 0b101;     /* Enable WDT interrupt in timeout of 0.5s */
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);      /* Enter power down mode */
  sleep_mode();
  WDTCR = _BV(WDE) | 0b110;                 /* Enable WDT reset in timeout of 1s */
}
//
//
/*-----------------------------------------------------------------------*/
/* Main                                                                  */
int main (void)
{ MCUSR = 0;
  WDTCR = _BV(WDE) | 0b110;                 /* Enable WDT reset in timeout of 1s */
  PORTB = 0b111011;                         /* Initialize port: - - H H H L H P */
  DDRB  = 0b111110;
  sei();
  unsigned int cp=0;                        //0 -     
  unsigned int N_Last;                      //      
  for (;;)
  { delay500();                             /* Delay 500ms in low power sleep mode */
    if (pf_mount(&Fs)) continue;            //Initialize FS -    FAT
  #ifdef SAVE_EEPR
  //     
    if (pf_opendir(&Dir, "")) break;        //Open sound file directory (root dir)
    while (!pf_readdir(&Dir, &Fno) && Fno.fname[0]) //  ,            NULL
    { if (!(Fno.fattrib & (AM_DIR|AM_HID))&& strstr(Fno.fname, ".WAV"))
      { K_Trek++;
      }
    }
//    eeprom_write_word(&ENtrek, K_Trek);     //      
  //      EEPROM
     N_Last=eeprom_read_word(&Eindex);      //      EEPROM
     if (N_Last>K_Trek) N_Last=0;           //   EEPROM   N_Trek - . =0
  #endif

    for (;;)
    { if (pf_opendir(&Dir, "")) break;      /*Open sound file directory (root dir) -  
                                              c     ""  ,
                                                 Dir.   pf_opendir()  FR_OK = 0 
                                              ,   .
                                            */
    //        
    #ifdef SAVE_EEPR
      if (cp==0)                            //     
      { while (cp<N_Last)                   //      
        { pf_readdir(&Dir, &Fno);
          cp++;
        }
        N_File=N_Last;
      }
      else N_File=0;
      cp=1;                                 //     
    #endif
    //
      while (!pf_readdir(&Dir, &Fno) && Fno.fname[0])/*Play all wav files in the dir - 
                                                  Dir (DirObject)
                                                  Fno (FileInfo)
                                                    ,   
                                               f_name[]  ,     FileInfo
                                                NULL    read index  
                                                       ,
                                                .
char* strstr (char *s1, char *s2)     s2   s1 ( "\0"  ).
    ,  NULL,    .
 s2     ,   s1.
                                            */
      { if (!(Fno.fattrib & (AM_DIR|AM_HID))&& strstr(Fno.fname, ".WAV"))
        { if (play(Fno.fname)) break;
        #ifdef SAVE_EEPR
         if (ButB>0)                      //     1  
         { if (pf_opendir(&Dir, "")) break;
           cp=0;
           if (N_File==0) N_File=K_Trek-1;
           else N_File--;
           while (cp<N_File)              //      
           { pf_readdir(&Dir, &Fno);
             cp++;
           }
           cp=1;                          //     
         }
         else N_File++;
        #endif
        }
      }
    }
  }
}
 
